/* $Id$
 * Copyright (C) 2005 Tommi Saviranta <wnd@iki.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* ifdef HAVE_CONFIG_H */

#include "common.h"
#include "ftc.h"
#include "protocol.h"
#include "progress.h"

#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>

#define DEBUG

#include <unistd.h>



void
print_error(const char *func, int ret, noiselevel noise)
{
	if (ret == -1) {
		if (noise != NOISE_QUIET) {
			fprintf(stderr, "\n%s: %s\n", func, strerror(errno));
		}
	} else {
		if (noise != NOISE_QUIET) {
			fprintf(stderr, "\n%s: failed! (%d bytes)\n",
					func, ret);
		}
	}
}



static int
adjust_blocksize(size_t *blocksize, int *bpu, struct progress *prog)
{
	int ops;

	ops = progress_get_ops(prog);

	if (ops > *blocksize * 2) {
		if (*blocksize == BUFMAXSIZE) {
			*bpu = ops / BUFMAXSIZE;
		} else {
			*blocksize *= 2;
			if (*blocksize > BUFMAXSIZE) {
				*blocksize = BUFMAXSIZE;
			}
		}
		return 1;
	} else if (ops < *blocksize / 2) {
		int nbpu;
		nbpu = ops / BUFMAXSIZE;
		if (nbpu > 1) {
			*bpu = nbpu;
		} else {
			*bpu = 1;
			*blocksize = ops;
		}
		return 2;
	}

	return 0;
}



static int
adjust_bpu(size_t block, size_t *blocksize, int *bpu, int tolerance,
		struct progress *prog)
{
	int ops;
	int new_bpu;

	ops = progress_get_ops(prog);
	if (ops == 0) {
		return 0;
	}
	new_bpu = ops / block;

	if (new_bpu > *bpu * tolerance) {
		*bpu *= 2;
		return 2;
	} else if (new_bpu < *bpu / tolerance) {
		*bpu /= 2;
		if (*bpu < 1) {
			*bpu = 1;
		}
		return 1;
	}
	return 0;
}



static char *
strip_path(const char *file)
{
	char *ptr;
	ptr = strrchr(file, (int) '/'); /* TODO support non-'/' */
	if (ptr == NULL) {
		return strdup(file);
	} else {
		return strdup(ptr + 1);
	}
}



int
common_put(int sock, const struct filelist files, int simple, noiselevel noise)
{
	uint8_t *buf;
	size_t blocksize;
	int bpu, bn;
	
	struct progress prog;
	char progtext[80];

	int i;
	int r;

	if (simple == 0) {
		r = protocol_put(sock, files, noise);
		if (r < 0) {
			printf("sending header failed!\n");
			return -1;
		}
	}

	blocksize = 1024;
	bpu = 1;
	bn = 0;

	buf = (uint8_t *) malloc(sizeof(uint8_t) * BUFMAXSIZE);
	if (buf == NULL) {
		if (noise != NOISE_QUIET) {
			fprintf(stderr, "Out of memory!\n");
		}
		return -1;
	}

	r = 0;
	for (i = 0; i < files.count && r != -1; i++) {
		FILE *file;
		uint32_t bytes;
		size_t block;

		file = fopen(files.name[i], "r");
		if (file == NULL) {
			print_error("fopen(\"r\")", -1, noise);
			r = -1;
			break;
		}

		progress_reset(&prog);
		prog.limit = files.size[i];

		if (noise != NOISE_QUIET) {
			char *t;
			t = malloc(strlen(files.name[i]) + 3);
			sprintf(t, "%s: ", files.name[i]);
			progress_set_stuff(&prog, t, 0, 0);
			free(t);
		}

		/* send stuff */
		bytes = 0;
		while (bytes < files.size[i]) {
			int up;
			size_t tblock;

			tblock = files.size[i] - bytes;
			if (tblock > blocksize) {
				tblock = blocksize;
			}
			block = fread(buf, 1, tblock, file);
			if (block != tblock) {
				if (ferror(file)) {
					print_error("fread()", -1, noise);
				}
				break;
			}
			bytes += block;

			r = send(sock, buf, block, 0);
			if (r != block) {
				print_error("send() data", r, noise);
				r = -1;
				break;
			}

			/* adjust blocksize */
			prog.state = bytes;
			up = adjust_blocksize(&blocksize, &bpu, &prog);

			/* time to update progress? */
			if (up == 0) {
				bn++;
				if (bn < bpu) {
					continue;
				}
			} else if (up == 2) {
				bn = 0;
				continue;
			} else {
				bn = 0;
			}

			if (noise != NOISE_QUIET) {
				prog.state = bytes;
				progress_print(&prog, progtext, 79);
				printf("\r%s", progtext);
				fflush(stdout);
			}
		}
		if (noise != NOISE_QUIET) {
			printf("\n");
		}

		fclose(file);
	}

	if (r == -1) {
		if (noise != NOISE_QUIET) {
			fprintf(stderr, "Sending \"%s\" failed!\n",
					files.name[i]);
		}
		return -1;
	}

	free(buf);

	return 0;
}



int
common_get(int sock, const char *filename, noiselevel noise)
{
	struct filelist *files;
	uint8_t *buf;
	size_t blocksize;
	int bpu, bn;

	struct progress prog;
	char progtext[80];

	int simple;
	int count;
	int i;
	int r;

	buf = (uint8_t *) malloc(sizeof(uint8_t) * BUFMAXSIZE);
	if (buf == NULL) {
		if (noise != NOISE_QUIET) {
			fprintf(stderr, "Out of memory!\n");
		}
		return -1;
	}

	if (filename == NULL) {
		files = protocol_get(sock, noise); /* TODO free files */
		if (files == NULL) {
			if (noise != NOISE_QUIET) {
				fprintf(stderr, "Reading header failed!\n");
			}
			return -1;
		}
		simple = 0;
		count = files->count;
	} else {
		simple = 1;
		count = 1;
	}

	blocksize = 1300; /* should somewhere be near MTU */
	bpu = 1;
	bn = 0;

	r = 0;
	for (i = 0; i < count && r != -1; i++) {
		FILE *file;
		char *name;
		uint32_t size;
		size_t block;
		uint32_t bytes;

		if (simple == 0) {
			name = strip_path(files->name[i]);
			size = files->size[i];
		} else {
			name = strdup(filename);
			size = 0;
		}

		file = fopen(name, "w");
		if (file == NULL) {
			print_error("fopen(\"w\")", -1, noise);
			free(name);
			r = -1;
			break;
		}

		progress_reset(&prog);
		prog.limit = size;

		if (noise != NOISE_QUIET) {
			char *t;
			t = malloc(strlen(name) + 3);
			sprintf(t, "%s: ", name);
			progress_set_stuff(&prog, t, 0, 0);
			free(t);
		}

		bytes = 0;
		while (bytes < size || simple == 1) {
			size_t tblock;

			tblock = size - bytes;
			if (tblock > blocksize) {
				tblock = blocksize;
			}
			block = recv(sock, buf, tblock, 0);
			if (block == -1) {
				print_error("recv() data", -1, noise);
				r = -1;
				break;
			} else if (block == 0) {
				if (simple == 1) {
					r = 0;
					break;
				} else {
					print_error("recv()", 0, noise);
					r = -1;
					break;
				}
			}
			bytes += block;

			r = fwrite(buf, 1, block, file);
			if (r != block) {
				print_error("fwrite()", -1, noise);
				r = -1;
				break;
			}
			/* extreme-case bpu adjust */
			adjust_bpu(block, &blocksize, &bpu, 10, &prog);
			bn++;
			if (bn < bpu) {
				continue;
			}

			bn = 0;

			/* adjust blocksize */
			prog.state = bytes;
			adjust_bpu(block, &blocksize, &bpu, 2, &prog);

			if (noise != NOISE_QUIET) {
				size_t nbs;
				
				prog.state = bytes;
				nbs = progress_print(&prog, progtext, 79);
				printf("\r%s", progtext);
				fflush(stdout);
			}
		}
		if (noise != NOISE_QUIET) {
			printf("\n");
		}

		fclose(file);
		free(name);
	}

	if (r == -1) {
		if (noise != NOISE_QUIET) {
			fprintf(stderr, "Receiving file failed!\n");
			return -1;
		}
	}

	free(buf);

	return r;
}
