/* $Id$
 * Copyright (C) 2005 Tommi Saviranta <wnd@iki.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* ifdef HAVE_CONFIG_H */

#include "ftc.h"
#include "server.h"
#include "client.h"

#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <getopt.h>

#include <sys/types.h>
#include <sys/stat.h>

#define DEFAULT_PORT	9951

typedef enum {
	ROLE_SERVER,
	ROLE_CLIENT,
	ROLE_SEND,
	ROLE_RECEIVE
} comprole;



int
main(int argc, char **argv)
{
	struct filelist files;
	noiselevel noise;
	comprole serve, trans;

	char *host;	/* computer to connect to */
	int port;	/* port to connect to */
	char *bind;	/* bind address TODO */
	int bindport;	/* bind port TODO */
	char *filename;	/* send/receive filename if single file mode */
	int simple;

	int usage;

	int ret;
	int i;

	usage = 0;

	filename = NULL;
	noise = NOISE_NORMAL;

	host = NULL;
	port = DEFAULT_PORT;
	bind = NULL;

	simple = 0;
	serve = ROLE_CLIENT;
	trans = ROLE_SEND;

	while (1) {
		int index, c;
		static struct option opts[] = {
			{ "help",	no_argument,		NULL,	'h' },

			{ "server",	no_argument,		NULL,	'S' },
			{ "client",	no_argument,		NULL,	'c' },

			{ "send",	no_argument,		NULL,	's' },
			{ "put",	no_argument,		NULL,	's' },
			{ "receive",	no_argument,		NULL,	'r' },
			{ "get",	no_argument,		NULL,	'r' },

			{ "target",	required_argument,	NULL,	't' },
			{ "host",	required_argument,	NULL,	't' },
			{ "port",	required_argument,	NULL,	'p' },

			{ "bind",	required_argument,	NULL,	'b' },
			{ "bind-port",	required_argument,	NULL,	'P' },

			{ "verbose",	no_argument,		NULL,	'v' },
			{ "quiet",	no_argument,		NULL,	'q' },

			{ "write-as",	required_argument,	NULL,	'w' },
			{ "save-as",	required_argument,	NULL,	'w' },

			{ "simple",	no_argument,		NULL,	'D' },
			{ "dummy",	no_argument,		NULL,	'D' },

			{ NULL,		0,			NULL,	'\0' }
		};

		
		c = getopt_long(argc, argv, "hScsrt:p:b:P:vqw:D", opts, &index);
		if (c == -1) {
			break;
		}

		switch (c) {
			case 'h':
				usage = 1;
				break;


			case 'S':
				serve = ROLE_SERVER;
				break;

			case 'c':
				serve = ROLE_CLIENT;
				break;


			case 's':
				trans = ROLE_SEND;
				break;

			case 'r':
				trans = ROLE_RECEIVE;
				break;


			case 't':
				if (host != NULL) {
					free(host);
				}
				host = strdup(optarg);
				if (host == NULL) {
					fprintf(stderr, "Out of memory!\n");
					return EXIT_FAILURE;
				}
				break;

			case 'p':
				port = atoi(optarg);
				break;


			case 'b':
				if (bind != NULL) {
					free(bind);
				}
				bind = strdup(optarg);
				if (bind == NULL) {
					fprintf(stderr, "Out of memory!\n");
					return EXIT_FAILURE;
				}
				break;

			case 'P':
				bindport = atoi(optarg);
				break;


			case 'v':
				noise = NOISE_VERBOSE;
				break;

			case 'q':
				noise = NOISE_QUIET;
				break;


			case 'w':
				if (filename != NULL) {
					free(filename);
				}
				filename = strdup(optarg);
				if (filename == NULL) {
					fprintf(stderr, "Out of memory!\n");
					return EXIT_FAILURE;
				}
				simple = 1;
				break;


			case 'D':
				simple = 1;
				break;


			case '?':
				usage = 1;
				break;

			default:
				fprintf(stderr, "Oops!\n");
				return EXIT_FAILURE;
		}
	}

	if (usage == 1) {
		printf(
"Usage: %s [OPTION ...] [FILE ...]\n\n"
"Options:\n"
"	-S	--server	Run as server\n"
"	-c	--client	Run as client\n"
"	-s	--send		Send file(s)\n"
"		--put		Send file(s)\n"
"	-r	--receive	Receive file(s)\n"
"		--get		Receive file(s)\n"
"	-t	--target	Address to connect to\n"
"		--host		Address to connect to\n"
"	-p	--port		Port to connect to\n"
"\n"
"	-b	--bind		Bind address\n"
/* "	-P	--bind-port	Port to bind\n" */
"\n"
"	-v	--verbose	Be verbose\n"
"	-q	--quiet		Be quiet\n"
"\n"
"	-w	--write-as	Save file as\n"
"		--save-as	Save file as\n"
"	-D	--simple	Simple mode (no protocol data)\n",
				argv[0]);
		return EXIT_SUCCESS;
	}

	files.count = (uint32_t) (argc - optind);

	if (trans == ROLE_SEND && files.count == 0) {
		if (noise != NOISE_QUIET) {
			fprintf(stderr, "no files to send\n");
		}
		return EXIT_FAILURE;
	}

	if (host == NULL && serve == ROLE_CLIENT) {
		if (noise != NOISE_QUIET) {
			fprintf(stderr, "no host given!\n");
		}
		return EXIT_FAILURE;
	}

	if (files.count != 0 && filename != NULL) {
		if (noise != NOISE_QUIET) {
			fprintf(stderr, "Can send only single files in "
					"simple/telnet mode!\n");
		}
		return EXIT_FAILURE;
	}

	files.name = (char **) malloc(sizeof(char *) * files.count);
	files.size = (uint32_t *) malloc(sizeof(uint32_t) * files.count);
	if (files.name == NULL || files.size == NULL) {
		printf("out of memory!\n"); /* TODO */
		exit(1);
	}
	for (i = 0; i < argc - optind; i++) {
		struct stat info;

		files.name[i] = strdup(argv[optind + i]);

		ret = stat(files.name[i], &info);
		if (ret == -1) {
			if (noise != NOISE_QUIET) {
				fprintf(stderr, "stat(): %d\n", errno);
			}
			return EXIT_FAILURE;
		}
		files.size[i] = (uint32_t) info.st_size;
		if (!S_ISREG(info.st_mode)) {
			if (noise != NOISE_QUIET) {
				fprintf(stderr, "\"%s\" is not regular file!\n",
						files.name[i]);
			}
			return EXIT_FAILURE;
		}

		/* TODO check if file can be read */

		if (noise == NOISE_VERBOSE) {
			printf("Going to send \"%s\" (%d bytes)\n",
					files.name[i],
					(int) files.size[i]);
		}
	}

	if (serve == ROLE_SERVER) {
		if (trans == ROLE_SEND) {
			ret = server_put(host, port, files,
					simple,
					NULL, /* TODO pass */
					noise);
		} else {
			ret = server_get(host, port,
					NULL, /* TODO pass */
					filename,
					noise);
		}
	} else {
		if (trans == ROLE_SEND) {
			ret = client_put(host, port, bind, bindport, files,
					simple,
					NULL, /* TODO pass */
					noise);
		} else {
			ret = client_get(host, port, bind, bindport,
					NULL, /* TODO pass */
					filename,
					noise);
		}
	}

	if (ret == 0) {
		return EXIT_SUCCESS;
	} else {
		return EXIT_FAILURE;
	}
}
