/* $Id$
 * Copyright (C) 2005 Tommi Saviranta <wnd@iki.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * This file was borrowed from miau-ng.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* ifdef HAVE_CONFIG_H */

#include "network.h"

#include <fcntl.h>
#include <errno.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#define PORT_STR_LEN	10


#define BIND_TRIES	32
#define SOCKTYPE	SOCK_STREAM
#define IPPROTO		IPPROTO_TCP



static void connect_resolve(new_connection_type *conn);
static int connect_next(new_connection_type *conn);
static void write_port_str(char *str, const int port);



int
network_connect_fast(const char *host, int port, const char *bind, int timeout)
{
	new_connection_type *conn;
	int sock;

	conn = network_connect_init(host, port, bind, timeout);
	if (conn == NULL) {
		return -1;
	}

	sock = 1;
	do {
		sock = network_connect(conn);
	} while (sock == -1);

	network_connect_free(conn);

	return sock;
} /* int network_connect_fast(const char *host, const int port,
		const char *bind, const int timeout) */



/**
 * network_connect_init:
 * @host	Address to connect
 * @port	Port to connect
 * @bind	Address to bind
 * @timeout	Timeout in ms
 *
 * Return value: #new_connection_type to be used with network_connect_next()
 *
 * TODO: timeout
 */
new_connection_type *
network_connect_init(const char *host, int port, const char *bind, int timeout)
{
	new_connection_type *conn;

	conn = malloc(sizeof(new_connection_type));
	conn->timeout = timeout;
	conn->state = CONNECTION_RESOLVE;
	conn->error_state = CONNECTION_FAILED_NOFAIL;

	conn->res = conn->res_iter = NULL;

	conn->host = strdup(host);
	conn->port = port;
	if (bind != NULL) {
		conn->bind = strdup(bind);
	} else {
		conn->bind = NULL;
	}

	conn->error[0] = '\0';
	conn->error[CONNECTION_ERR_LEN] = '\0';

	return conn;
} /* new_connection_type *network_connect_init(const char *host,
		const int port, const char *bind, const int timeout) */



int
network_connect(new_connection_type *conn)
{
	int sock;

	sock = -1;
	conn->error[0] = '\0';
	conn->error_state = CONNECTION_FAILED_NOFAIL;

	switch (conn->state) {
		case CONNECTION_RESOLVE:
			connect_resolve(conn);
			break;

		case CONNECTION_CONNECT:
			sock = connect_next(conn);
			break;
	}

	return sock;
} /* int network_connect(new_connection_type *conn) */



static void
connect_resolve(new_connection_type *conn)
{
	struct addrinfo	hints;
	char port_str[PORT_STR_LEN];
	
	int ret;
	
	bzero((void *) &hints, sizeof(hints));
	hints.ai_socktype = SOCKTYPE;
	hints.ai_family = PF_UNSPEC;
	hints.ai_protocol = IPPROTO;

	write_port_str(port_str, conn->port);
	ret = getaddrinfo(conn->host, port_str, &hints, &conn->res);
	if (ret != 0) {
		strncpy(conn->error, gai_strerror(ret), CONNECTION_ERR_LEN);
		conn->error_state = CONNECTION_FAILED_RESOLV;
	} else {
		conn->res_iter = conn->res;
		conn->state = CONNECTION_CONNECT;
	}
} /* static void connect_resolve(new_connection_type *conn) */



static int
connect_next(new_connection_type *conn)
{
	struct addrinfo *addr;
	int sock;
	int ret;

	addr = conn->res_iter;
	
	if (addr == NULL) {
		conn->state = CONNECTION_FAILED;
		return -1;
	}
	
	conn->res_iter = conn->res_iter->ai_next;

	/* Create socket. */
	sock = socket(addr->ai_family, addr->ai_socktype, addr->ai_protocol);
	if (sock == -1) {
		snprintf(conn->error, CONNECTION_ERR_LEN,
				"socket() failed"); /* TODO */
		conn->error_state = CONNECTION_FAILED_SOCKET;
		return -1;
	}
	
	/* Bind socket. */
	if (conn->bind != NULL) {
		/* TODO timeout */
		ret = network_bind_socket(sock, conn->bind, 0, conn->timeout);
		switch (ret) {
			case RET_BIND_BIND:
				snprintf(conn->error, CONNECTION_ERR_LEN,
						"couldn't bind");
				conn->error_state = CONNECTION_FAILED_BIND;
				return -1;
				break;
			case RET_BIND_RESOLVE:
				snprintf(conn->error, CONNECTION_ERR_LEN,
						"couldn't resolve");
				conn->error_state = CONNECTION_FAILED_BIND;
				return -1;
				break;
			case RET_BIND_TIMEOUT:
				strncpy(conn->error, strerror(errno),
						CONNECTION_ERR_LEN);
				conn->error_state = CONNECTION_FAILED_BIND;
				return -1;
				break;
		}
	}

	/* Connect. */
	ret = connect(sock, addr->ai_addr, addr->ai_addrlen);
	if (ret == -1) {
		conn->error_state = CONNECTION_FAILED_CONNECT;
		strncpy(conn->error, strerror(errno), CONNECTION_ERR_LEN);
		close(sock);
		sock = -1;
	} else {
		conn->state = CONNECTION_ESTABLISHED;
	}

	return sock;
} /* static int connect_next(new_connection_type *conn) */



/**
 * network_bind_socket:
 * @sock	Socket to bind
 * @host	Address to bind
 * @port	Port to bind. Set to 0 to ignore.
 * @timeout	Timeout in milliseconds
 *
 * Return value: On success, zero is returned. On error, 1 means bind failed,
 *	2 means address couldn't be resolved, 3 means bind was timed out.
 *
 * TODO: timeout
 */
int
network_bind_socket(int sock, const char *host, int port, int timeout)
{
	struct addrinfo *res;
	struct addrinfo *res_iter;

	struct addrinfo	hints;
	struct sockaddr addr;
	
	int ret;
	
	bzero((void *) &hints, sizeof(hints));
	hints.ai_socktype = SOCKTYPE;
	hints.ai_family = PF_UNSPEC;
	hints.ai_protocol = IPPROTO;

	if (port == 0) {
		ret = getaddrinfo(host, NULL, &hints, &res);
	} else {
		char port_str[PORT_STR_LEN];
		write_port_str(port_str, port);
		ret = getaddrinfo(host, port_str, &hints, &res);
	}
	if (ret == -1) {
		return RET_BIND_RESOLVE;
	}

	ret = -1;
	for (res_iter = res; res_iter != NULL && ret == -1;
			res_iter = res_iter->ai_next) {
		memcpy(&addr, res_iter->ai_addr, res_iter->ai_addrlen);
		ret = bind(sock, &addr, res_iter->ai_addrlen);
	}
	if (res != NULL) {
		freeaddrinfo(res);
	}
	if (ret == -1) {
		return RET_BIND_BIND;
	}
	
	return ret;
} /* int network_bind_socket(int sock, char *host, const int port,
		const int timeout) */



void
network_connect_free(new_connection_type *conn)
{
}



int
network_socket_new(void)
{
	int sock;
	sock = socket(AF_INET, SOCKTYPE, IPPROTO);
	return sock;
} /* int network_socket_new(void) */



/**
 * network_accept:
 * @sock	socket to accept
 * @sa		accepted connection
 *
 * Return value: -1 == error
 */
int
network_accept(int sock, struct sockaddr *sa)
{
	int ret;
	socklen_t size;
	struct sockaddr sa_l;

	if (sa == NULL) {
		sa = &sa_l;
	}
	size = sizeof(*sa);
	ret = accept(sock, sa, &size);

	return ret;
} /* int network_accept(int sock, struct sockaddr *sa) */



void
network_socket_close(int sock)
{
	close(sock);
} /* void network_socket_close(int sock) */



int
network_connect_get_next_address(new_connection_type *conn,
		const struct sockaddr **addr, size_t *addrlen)
{
	if (conn->res_iter == NULL) {
		return -1;
	}

	*addr = conn->res_iter->ai_addr;
	*addrlen = conn->res_iter->ai_addrlen;

	return 0;
} /* int network_connect_get_next_address(new_connection_type *conn,
		const struct sockaddr **addr, size_t *addrlen) */



int
network_resolv_sockaddr(char *dest, size_t n,
		struct sockaddr *sa, size_t addr_len)
{
	int ret;
	
	ret = getnameinfo(sa, addr_len, dest, n, NULL, 0, NI_NAMEREQD);

	return ret;
} /* int network_resolv_sockaddr(char *dest, const size_t n,
		struct sockaddr *sa, const size_t addr_len) */



static void
write_port_str(char *str, int port)
{
	snprintf(str, PORT_STR_LEN - 1, "%d", port);
	str[PORT_STR_LEN - 1] = '\0';
} /* static void write_port_str(char *str, const int port) */



int
network_socket_set_block(int sock)
{
	int flags;
	int r;

	flags = fcntl(sock, F_GETFL, 0);
	if (flags == -1) {
		return -1;
	}
	
	flags &= ~O_NONBLOCK;

	r = fcntl(sock, F_SETFL, flags);

	return r;
}
