/*
 * Copyright (C) 2005 Tommi Saviranta <wnd@iki.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* ifdef HAVE_CONFIG_H */

#include "progress.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>



void
progress_reset(struct progress *prog)
{
	struct timezone tz;

	prog->limit = prog->state = 0;

	bzero((void *) prog, sizeof(struct progress));
	gettimeofday(&prog->start_time, &tz);
	memcpy(&prog->last_time, &prog->start_time, sizeof(prog->start_time));
}



/*
int
progress_update(struct progress *prog)
{
}
*/



int
progress_get_ops(struct progress *prog)
{
	struct timeval now;
	struct timezone tz;
	double td;
	int bytes;
	int ops;

	gettimeofday(&now, &tz);

	td = 1000000.0 * (double) (now.tv_sec - prog->last_time.tv_sec) +
		(double) (now.tv_usec - prog->last_time.tv_usec);
	if (td == 0.0) { /* rather unlikely, but come up with something */
		td = 0.00000001;
	}
	bytes = (int) (prog->state - prog->last_state);

	ops = (int) ((double) bytes * 1000000.0 / td);

	return ops;
}



void
progress_set_stuff(struct progress *prog, const char *stuff,
		int pos, int trunc)
{
	if (prog->stuff != NULL) {
		free(prog->stuff);
	}
	prog->stuff = strdup(stuff);

	prog->stuff_pos = pos;
	prog->stuff_trunc = trunc;
}



static void
insert_stuff(struct progress *prog, char *buf, size_t size)
{
	size_t out_len;
	size_t stuff_len;
	size_t avail_len;

	out_len = strlen(buf);
	stuff_len = strlen(prog->stuff);
	avail_len = size - out_len;
	if (avail_len > stuff_len) {
		avail_len = stuff_len;
	}

	memmove(buf + avail_len, buf, out_len);
	buf[avail_len + out_len + 1] = '\0';
	buf[size - 1] = '\0';

	memcpy(buf, prog->stuff + stuff_len - avail_len, avail_len);
}



static void
print_eta(char *buf, size_t len, int seconds, int mode)
{
	int hr, min, sec;
	
	hr = seconds / 3600;
	min = (seconds - hr * 3600) / 60;
	sec = seconds % 60;

	switch (mode) {
		case 0:
			if (hr > 0) {
				snprintf(buf, len, "%d h %d m", hr, min);
			} else if (min > 0) {
				snprintf(buf, len, "%d m %d s", min, sec);
			} else {
				snprintf(buf, len, "%d s", sec);
			}
			break;

		case 1:
			if (hr > 0) {
				snprintf(buf, len, "%d:%02d:%02d",
						hr, min, sec);
			} else {
				snprintf(buf, len, "%d:%02d",
						min, sec);
			}
			break;

		case 2:
			if (hr > 0) {
				if (min + 3 > 60) {
					hr++;
				}
				min = (min + 3) / 6 * 10;
				snprintf(buf, len, "%d.%d hours",
						hr, min);
			} else if (min > 0) {
				if (sec > 30) {
					min++;
				}
				sec = (sec + 30) / 30 * 10;
				snprintf(buf, len, "%d.%d minutes",
						min, sec);
			} else {
				sec = (sec + 2) / 5 * 5;
				snprintf(buf, len, "%d seconds",
						sec);
			}
			break;

		default:
			buf[0] = '?';
			buf[1] = '\0';
			break;
	}

	buf[len - 1] = '\0';
}



static void
print_speed(char *buf, size_t len, int ops, int mode)
{
	switch (mode) {
		case 0:
			if (ops > (1 << 20)) {
				snprintf(buf, len, "%d MiB/s",
						(ops + (1 << 19)) >> 20);
			} else if (ops > (1 << 10)) {
				snprintf(buf, len, "%d KiB/s",
						(ops + (1 << 9)) >> 10);
			} else {
				snprintf(buf, len, "%d B/s", ops);
			}
			break;

		default:
			buf[0] = '?';
			buf[1] = '\0';
			break;
	}
}



int
progress_print(struct progress *prog, char *buf, size_t size)
{
#define TIMELEN 32
#define SPEEDLEN 16
	struct timeval now;
	struct timezone tz;
	int td;
	int bytes;
	int ops;

	gettimeofday(&now, &tz);

	td = 1000 * (int) (now.tv_sec - prog->last_time.tv_sec) +
		(int) (now.tv_usec - prog->last_time.tv_usec) / 1000;
	bytes = (int) (prog->state - prog->last_state);

	if (td != 0) {
		ops = (int) (bytes * 1000 / td);
	} else {
		ops = 0;
	}
	if (ops != 0) {
		char time[TIMELEN];
		char speed[SPEEDLEN];
		int sec;
		sec = (prog->limit - prog->state) / ops;
		print_eta(time, TIMELEN, sec, 2);
		print_speed(speed, SPEEDLEN, ops, 0);
		
		snprintf(buf, size, "%d bytes (%d %%) ETA %s, %s ",
			prog->state,
			prog->state * 100 / prog->limit,
			time,
			speed);
	} else {
		snprintf(buf, size, "%d bytes (%d %%)",
				prog->state,
				prog->state * 100 / prog->limit);
	}
	buf[size - 1] = '\0';

	if (prog->stuff != NULL) {
		insert_stuff(prog, buf, size);
	}

	prog->last_state = prog->state;
	memcpy(&prog->last_time, &now, sizeof(now));

	return ops;
}



/*
void
progress_set_limit_ascii(struct progress *progress, const char *limit)
{
	printf("TODO\n");
}
*/
