/* $Id$
 * Copyright (C) 2005 Tommi Saviranta <wnd@iki.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* ifdef HAVE_CONFIG_H */

#include "protocol.h"
#include "ftc.h"
#include "common.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>

#define BUFSIZE	65536

#define HAVE_NAME	0x01
#define HAVE_SIZE	0x02



static char empty_str[] = "";



static int
split_in_two(char *buf, char **t0, char **t1)
{
	char *t;

	*t0 = buf;
	*t1 = empty_str;

	t = strchr(buf, (int) ' ');
	if (t == NULL) {
		return 1;
	}

	*t = '\0';
	*t1 = t + 1;

	return 2;
}



static int
read_line(int sock, char *buf, size_t size)
{
	size_t os;
	int r;

	bzero(buf, size);
	for (os = 0; os < size - 1; os++) { /* make sure it's terminated */
		r = recv(sock, buf + os, 1, 0);
		if (r != 1) {
			return r;
		}
		if (buf[os] == '\n') {
			buf[os] = '\0';
#ifdef DEBUG
			printf("<< %s\n", buf);
#endif /* ifdef DEBUG */
			return 1;
		}
	}

	return 0;
}



struct filelist *
protocol_get(int sock, noiselevel noise)
{
	struct filelist *files;
	char buf[BUFSIZE];
	char *p0, *p1, *p2;
	int filecount;
	int i, ph;
	int r;

	if (noise == NOISE_VERBOSE) {
		printf("Getting header...\n");
	}

	r = read_line(sock, buf, BUFSIZE);
	if (r != 1) {
		print_error("recv() handshake", r, noise);
		return NULL;
	}
	r = split_in_two(buf, &p0, &p1);
	if (r != 2 || strcmp(p0, "FTC") != 0) {
		printf("handshake error: %s %s\n", p0, p1);
		return NULL;
	}
	printf("talking to \"%s\"\n", p1);

	/* header stuff */
	while (1) {
		int r;
		r = read_line(sock, buf, BUFSIZE);
		if (r != 1) {
			print_error("recv() header", r, noise);
			return NULL;
		}

		r = split_in_two(buf, &p0, &p1);

		if (strcmp(p0, "PASS") == 0) { /* TODO */
			r = split_in_two(p1, &p1, &p2);
			if (r == 2) {
				printf("pass: %s %s\n", p1, p2);
			} else {
				printf("bad pass: %s\n", p1);
			}
		}
		else if (strcmp(p0, "FILECOUNT") == 0) {
			if (r == 2) {
				filecount = atoi(p1); /* FIXME LFS */
			} else {
				printf("file count missing!\n"); /* TODO */
			}
		}
		else if (strcmp(p0, "HEADERDONE") == 0) {
			break;
		}
		else {
			printf("header: %s %s %s\n", p0, p1, p2); /* TODO */
		}
	}
	
	files = (struct filelist *) malloc(sizeof(struct filelist));
	if (files == NULL) {
		if (noise != NOISE_QUIET) {
			fprintf(stderr, "Out of memory!\n");
		}
		return NULL;
	}
	files->name = (char **) malloc(sizeof(char *) * filecount);
	files->size = (uint32_t *) malloc(sizeof(char *) * filecount);
	if (files->name == NULL || files->size == NULL) {
		printf("Out of memory!\n"); /* TODO */
		exit(1);
	}
	files->count = filecount;

	/* file list */
	i = -1;
	ph = 0;
	while (1) {
		int r;
		r = read_line(sock, buf, BUFSIZE);
		if (r < 1) {
			print_error("recv() file list", r, noise);
			/* TODO free file list */
			return NULL;
		}

		r = split_in_two(buf, &p0, &p1);

		if (strcmp(p0, "FILE") == 0 && r == 2 &&
				(ph == 0 || (ph & HAVE_SIZE))) {
			i++;
			if (i == filecount) {
				printf("too many files!\n"); /* TODO */
				exit(1);
			}
			files->name[i] = strdup(p1);
			ph = HAVE_NAME;
		}
		else if (strcmp(p0, "SIZE") == 0 && r == 2 &&
				(ph & HAVE_NAME)) {
			files->size[i] = atoi(p1); /* TODO LFS */
			ph |= HAVE_SIZE;
		}
		else if (strcmp(p0, "UMODE") == 0 && r == 2 &&
				(ph & HAVE_NAME)) {
			printf("umode %s\n", p1);
		}
		else if (strcmp(p0, "SUM") == 0 && r == 2 &&
				(ph & HAVE_NAME)) {
			r = split_in_two(p1, &p1, &p2);
			if (r == 2) {
				printf("sum %s %s\n", p1, p1);
			} else {
				printf("bad sum: %s", p1);
			}
		}
		else if (strcmp(p0, "FILESDONE") == 0) {
			printf("expected %d files, got %d\n", filecount, i + 1);
			/* TODO */
			break;
		}
		else {
			printf("file info: %s %s %s\n", p0, p1, p2);
		}
	}

	/* other stuff */
	while (1) {
		r = read_line(sock, buf, BUFSIZE);
		if (r != 1) {
			print_error("recv() other info", r, noise);
			return NULL;
		}

		r = split_in_two(buf, &p0, &p1);

		if (strcmp(p0, "ALLDONE") == 0) {
			break;
		}
		else {
			printf("other info: %s %s\n", p0, p1);
		}
	}

	return files;
}



int
protocol_put(int sock, struct filelist files, noiselevel noise)
{
	char buf[BUFSIZE];
	int t, r;
	int i;

	if (noise == NOISE_VERBOSE) {
		printf("Sending header...\n");
	}

	bzero(buf, BUFSIZE);

	/* header */
	snprintf(buf, BUFSIZE,
			"FTC ftv %s\n"
			"FILECOUNT %d\n"
			"HEADERDONE\n",
			VERSION,
			files.count);
	buf[BUFSIZE - 1] = '\0';
	t = strlen(buf);
	r = send(sock, buf, t, 0);
	if (r != t) {
		print_error("send() header", r, noise);
		return -1;
	}

	/* files */
	for (i = 0; i < files.count; i++) {
		snprintf(buf, BUFSIZE,
				"FILE %s\n"
				"SIZE %d\n",
				files.name[i],
				files.size[i]);
		buf[BUFSIZE - 1] = '\0';
		t = strlen(buf);
		r = send(sock, buf, t, 0);
		if (r != t) {
			print_error("send() files", r, noise);
			return -1;
		}
	}
	snprintf(buf, BUFSIZE, "FILESDONE\n");
	buf[BUFSIZE - 1] = '\0';
	t = strlen(buf);
	r = send(sock, buf, t, 0);
	if (r != t) {
		print_error("send() files", r, noise);
		return -1;
	}

	snprintf(buf, BUFSIZE, "ALLDONE\n");
	buf[BUFSIZE - 1] = '\0';
	t = strlen(buf);
	r = send(sock, buf, t, 0);
	if (r != t) {
		print_error("send() header end", r, noise);
		return -1;
	}

	return 0;
}
